from .exceptions import LoginFailure, NonPublicFunctionError, UploadError, QuestionError
from .types.Question import Question
from .types.Subject import Subject
from .types.Notif import Notif
from .types.User import User
from .types.List import List # too much troll, (me a couple months later) what the actual fuck did i mean with this
from pathlib import Path
import requests, json, platform

class Session:
	def __init__(self, token):
		self.token = token["auth_token"]

	def get_subjects(self):
		resp = requests.get("https://api.wrts.nl/api/v3/subjects",headers={"X-Auth-Token": self.token}).json()
		return (Subject(x,self) for x in resp["subjects"])

	def upload(self, path: Path, mimetype="image/png"):
		with open(path, "rb") as f:
			data = f.read()

		headers = {"X-Auth-Token": self.token}
		headers.update({"Content-type":mimetype})

		resp = requests.get("https://api.wrts.nl/api/v3/qna/questions/presigned_image_url", headers={"X-Auth-Token": self.token}).json()
		stat = requests.put(resp["signed_url"],headers=headers,data=data).status_code

		if stat == 200:
			return {"file_name": path.parts[-1], "image": resp["signed_url"].split("?")[0]}
		else:
			raise UploadError(f"Failed to upload {path}")

	def get_notifs(self, page, per_page=10):
		resp = requests.get(f"https://api.wrts.nl/api/v3/users/notifications?page={page}&per_page={per_page}", headers={"x-auth-token": self.token}).json()
		return resp["total_count"], (Notif(x,self) for x in resp["notifications"])

	def get_questions(self):
		resp = requests.get("https://api.wrts.nl/api/v3/public/qna/questions", headers={"x-auth-token": self.token}).json()
		return resp["total_count"], (Question(x["id"],self) for x in resp["results"])

	def get_question(self, id):
		return Question(id, self)

	def post_question(self, contents, subject, topic=None, attachments=[]):
		data = {"contents": contents, "subject_id": subject.id, "qna_attachments_attributes": attachments}
		if not topic == None:  data["topic_id"] = topic.id

		resp = requests.post("https://api.wrts.nl/api/v3/qna/questions",headers={"x-auth-token": self.token},json=json.dumps({"qna_question":data})).json()
		if "success" in resp:
			raise QuestionError(resp["error"])
		return self.get_question(resp["qna_question"]["id"])

	def get_self(self): # return user
		resp = requests.get("https://api.wrts.nl/api/v3/get_user_data", headers={"X-Auth-Token": self.token}).json()
		return User(resp["username"], self)

	def get_subject(self, id):
		for subject in self.get_subjects():
			if subject.id == id:
				return subject
			

class PublicSession(Session):
	def __init__(self):
		super().__init__("")

	def get_subjects(self):
		raise NonPublicFunctionError("This function is not available for public use, an inlog is required!")
	
	def upload(self, path, mimetype="image/png"):
		raise NonPublicFunctionError("This function is not available for public use, an inlog is required!")
	
	def get_notifs(self, page, per_page=10):
		raise NonPublicFunctionError("This function is not available for public use, an inlog is required!")
	
	def get_self(self):
		raise NonPublicFunctionError("This function is not available for public use, an inlog is required!")
	
	def post_question(self, contents, subject, topic=None, attachments=[]):
		raise NonPublicFunctionError("This function is not available for public use, an inlog is required!")




from .User import User
import requests

class NotifError(Exception):
	pass

class Notif:
	def __init__(self, obj, session):
		self.session = session
		self.id = obj["id"]
		self.created_at = obj["created_at"]
		self.icon = obj["icon"]
		self.image = obj["image"] # ???
		self.is_recent = obj["is_recent"]
		self.landing_url = obj["landing_url"]
		self.message = obj["message"]
		self.read = obj["read"]
		self.retrieved_at = obj["retrieved_at"] # aka read_at i think
		self.creator = User(obj["creator"]["public_profile_name"],session)
	def read(self):
		resp = requests.patch(f"https://api.wrts.nl/api/v3/users/notifications/{self.id}",headers={"x-auth-token": self.session.token}).json()
		if not resp["success"]:
			raise NotifError(resp["error"]+"..... bro how did you mess up reading a notification") # bro how does one even mess that up
		return True



from .Session import *
from .exceptions import *
import requests, json, os, time

def get_token(email, password):
	resp = requests.post("https://api.wrts.nl/api/v3/auth/get_token", json={"email": email, "password": password}).json()
	return resp


def save_token(email, password, path):
	with open(path, "w+") as f:
		resp = get_token(email, password)
		f.write(json.dumps(resp))
	return resp

def load_from_file(path):
	try:
		with open(path) as f:
			data = json.loads(f.read())
			return data
	except json.decoder.JSONDecodeError:
		return {"success": False}


def login(email, password, path=".WRTS"):
	try:
		token = load_from_file(path)
		if time.time() > token["expires_at"]:
			raise FileNotFoundError("Sussy")

	except FileNotFoundError:
		token = save_token(email, password, path)

	if not token["success"]:
		raise LoginFailure(token["info"])
	return Session(token)



# TODO: Handle these exceptions properly

class LoginFailure(Exception):
	pass

class QuestionError(Exception):
	pass

class UploadError(Exception):
	pass

class UserError(Exception):
	pass

class NonPublicFunctionError(Exception):
	pass


#from wrts.types.List import List
from wrts.exceptions import UserError
import requests, re

class Rank:
	def __init__(self, text, logo):
		self.rank = re.search("[\w ]*(?!\()", text).group(0)
		self.points = int(text.split(" ")[-2][1:])
		self.logo = logo

class ProfileImage:
	def __init__(self, obj):
		self.url = obj["image_url"]
		self.default_color = obj["profile_color"]
		self.default_letter = obj["profile_letter"]

		self.default_pfp = self.url == None # If no URL is given, it must be a default profile picture

class User:
	def __init__(self, path, session):
		resp = requests.get("https://api.wrts.nl/api/v3/public/users/"+path, headers={"x-auth-token": session.token}).json()
		if "success" in resp: raise UserError(resp["error"])
		obj = resp["user"]
		self.first_name = obj["first_name"]
		self.id = obj["id"]
		self.path = path
		self.is_self = obj["is_own_profile"] # why would anybody need this
		self.profile_image = ProfileImage(obj["profile_image"])
		self.profile_image_path = obj["profile_image_path"] # BRO, WHY, profile_image.image_url EXISTS!!!
		self.rank = Rank(obj["qna_rank_and_points_display"], obj["qna_rank_logo"])
		self.tutor = obj["tutor"] # <-- aka self.gay = obj["gay"]
		self.session = session

	def get_lists(self):
		List = __import__("wrts.types.List").List # I know, it's a very bad hack however if it works, it works
		resp = requests.get(f"https://api.wrts.nl/api/v3/public/users/{self.path}/practiceable_items", headers={"X-Auth-Token": self.session.token}).json()
		return (List(o["id"],self.session) for o in resp)




class FLAGGING_TYPES:
	INAPPROPRIATE = 1
	CONTAINS_LINK = 2
	WRONG_SUBJECT = 3
	ADVERTISING   = 4
	PERSONAL_INFO = 5

class PERFORMANCE_ORDER:
	WRONG = 0
	MOSTLY_WRONG = 1
	SOMETIMES_WRONG = 2
	MOSTLY_RIGHT = 3
	GOOD = 4
	NOT_PRACTICED = 5

class EXERCISE_TYPES:
	LEARN = 'learn'
	DICTATE = 'dictate'
	TEST = 'full_word'
	HINTS = 'hints'
	MENTAL = 'in_your_mind'
	MULTIPLE_CHOICE = 'multiple_choice'

class BATTLE_EXERCISE_TYPES:
	TEST = 'full_word'
	HINTS = 'hints'
	MULTIPLE_CHOICE = 'multiple_choice'
	SPELLING = 'timed'

class ANSWER_TYPES:
	FULL = 'full_word'
	LEARN = 'learn'
	MULTIPLE_CHOICE = 'multiple_choice'
	HINTED = 'hints'


from wrts.exceptions import NonPublicFunctionError
from wrts.types.User import User
from wrts.types.Subject import Subject, Topic
from wrts.enums import ANSWER_TYPES, EXERCISE_TYPES
from datetime import datetime
import requests, random

class PracticeSession:
    def __init__(self, obj, session):
        self.session = session

        self.id = obj["id"]
        self.list_id = list(obj["list_languages"].keys())[0]
        self.words = {o["id"]: o["data"] for o in obj["words"]}
        self.special_chars = obj["special_characters"]
        self.extype = obj["exercise_type_code"]
        self.answer_url = obj["answer_url"]
        self.onboarding = obj["show_onboarding"]
        self.all_config = obj["all_configuration"]
        self.legit_config = obj["legit_configuration_values"] # ???
        self.simplified_attemps = obj["simplified_attempts"] #??????/
        self.word_queue = obj["word_queue"]
        self.logic_params = obj["logic_parameters"] # ???????????????????????????
        self.typochecker = obj["allow_typochecker"]
        self.langs = obj["list_languages"][self.list_id]
        self.config = obj["configuration"]

        self.word_queue = []
        self.progress = 0
        self.finished = False
        self.answer_locale = self.config[0]["settings"][0]["value"]
        self.question_locale = ""

        for i, lang in enumerate(self.langs):
            if not lang["locale"] == self.answer_locale:
                self.question_locale = lang["locale"]

        if self.config[0]["settings"][1]["value"]:
            random.shuffle(self.words)

        match self.extype:
            case EXERCISE_TYPES.LEARN:
                self.generate_learn_queue()
            case EXERCISE_TYPES.TEST:
                self.generate_test_queue()

    def push_to_queue(self, display_type: str, word_id: str, roundnr=1):
        if word_id=="": return # dumb hack for generate_learn_queue
        self.word_queue.append({
            "answer_locale": self.answer_locale,
            "correct": None,
            "display_type": display_type,
            "list_id": self.list_id,
            "question_locale": self.question_locale,
            "round_nr": roundnr,
            "word_id": word_id
        })

    def generate_test_queue(self):
        for wordid in self.words:
            self.push_to_queue(ANSWER_TYPES.FULL, wordid)

    def generate_learn_queue(self):
        words_keys = list(self.words.keys())
        words_list = [(words_keys[i], words_keys[i+1]) for i in range(len(words_keys)//2)]
        if not len(words_keys)//2*2 == len(words_keys): words_list.append((words_keys[-1],""))

        for pair in words_list:
            self.push_to_queue(ANSWER_TYPES.LEARN, pair[0])
            self.push_to_queue(ANSWER_TYPES.LEARN, pair[1])

            self.push_to_queue(ANSWER_TYPES.MULTIPLE_CHOICE, pair[0])
            self.push_to_queue(ANSWER_TYPES.MULTIPLE_CHOICE, pair[1])
            
            self.push_to_queue(ANSWER_TYPES.HINTED, pair[0])
            self.push_to_queue(ANSWER_TYPES.HINTED, pair[1])
            
            self.push_to_queue(ANSWER_TYPES.FULL, pair[0])
            self.push_to_queue(ANSWER_TYPES.FULL, pair[1])



    def get_question(self):
        wid = self.word_queue[self.progress]["word_id"]
        
        retval = None
        match self.word_queue[self.progress]["display_type"]:
            case ANSWER_TYPES.LEARN:
                retval = (self.words[wid][self.question_locale], self.words[wid][self.answer_locale])

            case ANSWER_TYPES.MULTIPLE_CHOICE:
                options = [self.words[wid][self.answer_locale]]
                while len(options) < 4:
                    candidate = self.words[random.choice(list(self.words.keys()))][self.answer_locale]
                    if not candidate == self.words[wid][self.answer_locale]:
                        options.append(candidate)
                
                retval = (self.words[wid][self.question_locale], options)
            case ANSWER_TYPES.HINTED:
                retval = (self.words[wid][self.question_locale], self.words[wid][self.answer_locale][0])

            case ANSWER_TYPES.FULL:
                retval = self.words[wid][self.question_locale]

        return retval, self.word_queue[self.progress]["display_type"]

    def answer(self, answer):
        if self.progress == len(self.word_queue)-1:
            self.finished = True

        wid = self.word_queue[self.progress]["word_id"]

        if self.words[wid][self.answer_locale] == answer:
            self.word_queue[self.progress]["correct"] = True
        else:
            self.word_queue[self.progress]["correct"] = False

        if self.extype == EXERCISE_TYPES.LEARN and not self.word_queue[self.progress]["display_type"] == ANSWER_TYPES.LEARN and not self.word_queue[self.progress]["correct"]:
            display_type = self.word_queue[self.progress]["display_type"]
            roundnr = self.word_queue[self.progress]["round_nr"]
            self.push_to_queue(display_type, wid, roundnr)
            self.finished = False

        req = {
            "answer_locale": self.answer_locale,
            "correct_answer": self.words[wid][self.answer_locale],
            "display_type": self.word_queue[self.progress]["display_type"],
            "is_answer_correct": self.word_queue[self.progress]["correct"],
            "is_exercise_finished": self.finished,
            "marked_as_correct_answer": None,
            "provided_answer": answer,
            "question_locale": self.question_locale,
            "round_nr": 1, # handle this correctly next time
            "typochecker_accepted": None,
            "typochecker_shown": False, # These aswell
            "word_id": wid,
            "word_queue": self.word_queue
        }

        if req["display_type"] == ANSWER_TYPES.LEARN:
            del req["correct_answer"]
            req["is_answer_correct"] = True
            self.word_queue[self.progress]["correct"] = True
            req["word_queue"] = self.word_queue

            

        self.progress += 1

        resp = requests.post(self.answer_url, headers={"x-auth-token": self.session.token}, json=req).json()
        return req["is_answer_correct"], resp["success"]
        
        

class Result:
    def __init__(self, obj):
        self.id = obj["id"]
        self.grade = obj["grade"]
        self.started = datetime.strptime(obj["started_at"], "%Y-%m-%dT%H:%M:%S.000Z")
        self.finished = datetime.strptime(obj["finished_at"], "%Y-%m-%dT%H:%M:%S.000Z")
        self.exercise_type = obj["exercise_type_code"]
        self.errors = obj["wrong_answers_count"]
        self.corrects = obj["correct_answers_count"]
        self.accuracy = obj["correctness_percentage"]
        self.practiced_percentage = obj["practiced_words_percentage"]
        self.practiced = obj["words_practiced"]
        self.list_length = obj["words_in_lists"]
        self.first_attempts = (obj["first_attempts_correctness"]["correct"], obj["first_attempts_correctness"]["incorrect"])

class List:
    def __init__(self, id, session):
        obj = requests.get(f"https://api.wrts.nl/api/v3/public/lists/{id}").json()
        
        self.id = id
        self.session = session

        self.title = obj["title"]
        self.creator = User(obj["creator"]["public_profile_url"], session)
        self.times_practiced = obj["times_practiced"]
        self.shared = obj["shared"]
        self.deleted = obj["deleted"]
        self.rated_words = obj["words_with_performance"] # properly parse this one
        self.book = obj["book"] # same goes for this one
        self.related_topics = (Topic(top, None, None) for top in obj["related_topics"]) # this too
        self.paused_exercise = obj["paused_exercise"] # aswell
        self.status = obj["status"]
        self.upgrade_required = obj['needs_upgrade']
        self.minrole = obj["min_required_role"]
        self.word_count = obj["word_count"]
        self.upgrade_info = obj["upgrade_info"]
        self.related_type = obj["related_topics_type"]
        self.chapter = obj["chapter"] # parse
        self.subjects = (Subject(sub) for sub in obj["subjects"])
        self.prioritylang = obj["prioritized_language"]
        self.locales = obj["locales"]

    def get_results(self):
        if self.session.token == "":
            raise NonPublicFunctionError("This function is not available for public use, an inlog is required!")
        resp = requests.get(f"https://api.wrts.nl/api/v3/results?list_id={self.id}", headers={"x-auth-token": self.session.token}).json()["results"]
        return (Result(res) for res in resp)

    def practice(self, extype, id=None, selected_words=[]):
        if self.session.token == "":
            raise NonPublicFunctionError("This function is not available for public use, an inlog is required!")
        req = {
            "exercise_type_code": extype,
            "id": id,
            "list_ids": [self.id],
            "selected_words": selected_words
        }

        resp = requests.post("https://api.wrts.nl/api/v3/exercises", headers={"x-auth-token": self.session.token}, json=req).json()
        return PracticeSession(resp, self.session)


from wrts.exceptions import NonPublicFunctionError, QuestionError
from .Subject import *
from .User import *
import json, requests

class Answer:
	def __init__(self, obj, session):
		self.session = session
		self.body = obj["body"]
		self.can_edit = obj["can_edit"]
		self.can_flag = obj["can_flag"]
		self.correct = obj["correct"]
		self.id = obj["id"]
		self.is_flagged = obj["is_flagged"]
		self.is_own_answer = obj["is_own_answer"]
		self.attachments = obj["qna_attachments"]
		self.upvoted_by_self = obj["is_upvoted"]
		self.user = User(obj["user"]["username"], self.session)
	def vote(self):
		result = requests.post(f"https://api.wrts.nl/api/v3/qna/answers/{self.id}/votes", headers={"x-auth-token": self.session.token}).text
		if self.upvoted_by_self:
			self.votes -= 1
			self.upvoted_by_self = False
		else:
			self.votes += 1
			self.upvoted_by_self = True

class Question:
	def __init__(self, id, session):
		obj = requests.get(f"https://api.wrts.nl/api/v3/public/qna/questions/{id}", headers={"X-auth-token": session.token}).json()
		if obj["qna_question"] == None: raise QuestionError("Question not found")
		obj = obj["qna_question"]
		self.session = session
		self.body = obj["body"]
		self.can_answer = obj["can_answer"]
		self.can_edit = obj["can_edit"]
		self.flag = obj["can_flag"]
		self.contents = obj["contents"]
		self.creation = obj["created_at"]
		self.id = id
		self.is_flagged = obj["is_flagged"]
		self.answers = (Answer(a,self.session) for a in obj["other_qna_answers"])
		self.tutor_answers = (Answer(a,self.session) for a in obj["tutor_qna_answers"])
		self.attachments = obj["qna_attachments"] # do not forget to parse this later
		self.moderated = obj["requires_forced_moderation"]
		self.title = obj["title"]
		#self.subject = Subject(obj["subject"],token)
		self.user = User(obj["user"]["username"],self.session)

		subjects = requests.get("https://api.wrts.nl/api/v3/subjects",headers={"x-auth-token": self.session.token}).json()["subjects"]
		for sub in subjects:
			if sub["id"] == obj["subject"]["id"]:
				self.subject = Subject(sub,session)
				break
		if not obj["topic"] == None:
			for topic in self.subject.topics:
				if topic.id == obj["topic"]["id"]:
					self.topic = topic
					break
	def answer(self, body, attachments=[]):
		if self.session.token == "":
			raise NonPublicFunctionError("This function is not available for public use, an inlog is required!")
		
		resp = requests.post(f"https://api.wrts.nl/api/v3/qna/questions/{self.id}/answers", json={"qna_answer":{"body": body, "qna_attachments_attributes": attachments}}, headers={"X-Auth-Token": self.session.token}).json()
		return Answer(resp["qna_answer"]["id"], self.session)
	def get_related_questions(self):
		resp = requests.get(f"https://api.wrts.nl/api/v3/public/qna/questions/{self.id}/related_questions", headers={"x-auth-token":self.session.token}).json()
		return resp["label"], resp["total_count"], (Question(o["id"],self.session) for o in resp["qna_questions"])

	def report(self, reason):
		if self.session.token == "":
			raise NonPublicFunctionError("This function is not available for public use, an inlog is required!")
		resp = requests.post("https://api.wrts.nl/api/v3/qna/flagged_questions", headers={"x-auth-token"}, json={"qna_question_id": self.id, "qna_question_flagging_reason": reason}).json
		return resp

